/********************************************************************
 * FILE: gomo.c
 * AUTHOR: Fabian Buske
 * CREATE DATE: 18/06/2008
 * PROJECT: MEME suite
 * COPYRIGHT: 2008, UQ
 *
 * GOMO is an implementation of the algorithm described in
 * "Associating transcription factor binding site motifs with target
 * Go terms and target genes"
 * authors: Mikael Boden and Timothy L. Bailey
 * published: Nucl. Acids Res (2008)
 *
 ********************************************************************/

#include <assert.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <float.h>
#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
// matrix.h has some strange requirement that it be defined before arrays.h
// as one of the .h files below includes arrays.h (I have no idea which) 
// matrix.h needs to stay at the top...
#include "matrix.h"
#include "alphabet.h"
#include "array-list.h"
#include "binary-search.h"
#include "cisml.h"
#include "config.h"
#include "dir.h"
#include "fasta-io.h"
#include "hash_table.h"
#include "io.h"
#include "linked-list.h"
#include "meme-io.h"
#include "merger.h"
#include "projrel.h"
#include "qvalue.h"
#include "ranksum_test.h"
#include "read_csv.h"
#include "read_tamo.h"
#include "simple-getopt.h"
#include "string-list.h"
#include "xml-util.h"

#define min(a,b)      (a<b)?a:b
#define max(a,b)      (a<b)?b:a
#define bool2txt(b)   (b) ?"TRUE":"FALSE"

/*********************************************************************************
 * Constants 
 ********************************************************************************/
#define DEFAULT_Q_VALUE_THRESHOLD 0.05  /* Q-value threshold */
#define BIN_ALLOC_CHUNK 10 /* number of tuples to increment bins by */
#define DEFAULT_SHUFFLES 10
#define DEFAULT_MIN_GENE_COUNT 1
const char *program_name = "gomo"; /* the program name */
static char *default_output_dirname = "gomo_out";  /* default name of output */
static const char *XML_FILENAME = "gomo.xml";
static const char *HTML_STYLESHEET = "gomo.xsl";
static const char *HTML_FILENAME = "gomo.html";
//for get_gomap_path, which is not used. 
/*********************************************************************************
 * Globals
 ********************************************************************************/
VERBOSE_T verbosity = NORMAL_VERBOSE;
BOOLEAN_T status = TRUE;

/*************************************************************************
 * gomo_dts()
 *************************************************************************/
static char *gomo_dts = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n"
    "<?xml-stylesheet type=\"text/xsl\" href=\"gomo.xsl\"?>\n"
    "<!-- Document definition -->\n"
    "<!DOCTYPE gomo[\n"
    "\t<!ELEMENT gomo (program, motif*)>\n"
    "\t<!ATTLIST gomo xmlns:xsi CDATA #IMPLIED\n"
    "\t\t\t version CDATA #REQUIRED\n"
    "\t\t\t release CDATA #REQUIRED\n"
    "\t\t\t>\n"
    "\t\t<!ELEMENT program (gomapfile, seqscorefile+)>\n"
    "\t\t<!ATTLIST program\n"
    "\t\t\t\t name CDATA #REQUIRED\n"
    "\t\t\t\t cmd CDATA #REQUIRED\n"
    "\t\t\t\t outdir CDATA #REQUIRED\n"
    "\t\t\t\t clobber CDATA #REQUIRED\n"
    "\t\t\t\t text_only CDATA #REQUIRED\n"
    "\t\t\t\t use_e_values CDATA #REQUIRED\n"
    "\t\t\t\t score_e_thresh CDATA #REQUIRED\n"
    "\t\t\t\t min_gene_count CDATA #REQUIRED\n"
    "\t\t\t\t motifs CDATA #IMPLIED\n"
    "\t\t\t\t shuffle_scores CDATA #REQUIRED\n"
    "\t\t\t\t q_threshold CDATA #REQUIRED\n"
    "\t\t\t\t>\n"
    "\t\t\t<!ELEMENT gomapfile EMPTY>\n"
    "\t\t\t<!ATTLIST gomapfile path CDATA #REQUIRED >\n"
    "\t\t\t<!ELEMENT seqscorefile EMPTY>\n"
    "\t\t\t<!ATTLIST seqscorefile path CDATA #REQUIRED >\n"
    "\t\t<!ELEMENT motif (goterm*)>\n"
    "\t\t<!ATTLIST motif id CDATA #REQUIRED >\n"
    "\t\t\t<!ELEMENT goterm EMPTY>\n"
    "\t\t\t<!ATTLIST goterm\n"
    "\t\t\t\t\t id CDATA #REQUIRED\n"
    "\t\t\t\t\t score CDATA #REQUIRED\n"
    "\t\t\t\t\t pvalue CDATA #REQUIRED\n"
    "\t\t\t\t\t qvalue CDATA #REQUIRED\n"
    "\t\t\t\t\t description CDATA #REQUIRED\n"
    "\t\t\t\t\t>\n"
    "]>\n";

/***********************************************************************************
 * Data types
 **********************************************************************************/
/*
 * seq mapping
 * Provides a mapping from one sequence to another
 * allowing shuffling of labels on sequence scores.
 */
typedef struct seq_mapping SHUFFLER_T;
struct seq_mapping {
  int *map;           // list that contains a mapping for all loaded sequences
                      // provides a conversion from one sequence to another
                      // a sequential numbering(map[i] = i) provides the
                      // identity mapping.
  int map_length;     // the number of entries in map
  int *rank;          // list that contains lookup for go map sequences
                      // generated using the seq_map to map a gomap sequence
                      // to a rank in the sorted score list
  int rank_length;    // the number of entries in rank
};

/*
 * ama input
 */
typedef struct ama_score AMA_SCORE_T;
struct ama_score {
  int index; // the load number of the sequence 
  double score; // the score for a sequence id, it will be referred to by the index
};

/*
 * ama list
 */
typedef struct ama_list AMA_LIST_T;
struct ama_list {
  AMA_SCORE_T *list;
  int length;
};

/* gomo result */
typedef struct gomo_result GOMO_RESULT_T;
struct gomo_result {
  char* go_id; // go term 
  double gomo_score; // geometric mean of rank sum p-values
  double empirical_pvalue; // score for go term
};

/*
 * contains the description of the go term
 * and the index of that go term in all mappings
 */
typedef struct goinfo_t GOINFO_T;
struct goinfo_t {
  char *desc;
  int index;
};

/*
 * A gene list contains a list of genes.
 * The size is a value that is larger than zero
 * and is the number of elements in the genes array.
 * The strings pointed to by 'ids' are not allocated
 * by this gene list and so should not be deallocated
 * when the gene list is deallocated.
 */
typedef struct seq_list_t SEQ_LIST_T;
struct seq_list_t {
  int length;
  int *seqnums;
};

/*
 * A tuple contains the multi-species GOMO p-value associated with 
 * a given GO term and the counts of Null model items that have been 
 * better. Note: the go_id is not copied on creation so it
 * should not be deallocated on destruction.
 */
typedef struct tuple_t TUPLE_T;
struct tuple_t {
  char *go_id;		// the GO term ID
  double gomo_score;	// the gomo score of the term 
  int counts; // number of counts
};

/*
 * Each GO term has a number of genes associated with it, called
 * its "gene_count".  The BIN_T is designed to contain
 * all of the tuples (goid, pvalue, counts) entries for
 * all GO terms within the given gene count range.  The tuples are
 * stored in an array that always has an allocated size that is a 
 * multiple of BIN_ALLOC_CHUNK. 
 */
typedef struct bin_t BIN_T;
struct bin_t {
  int low_gene_count; // low bound on gene count (inclusive)
  int high_gene_count; // high bound on gene count (inclusive)
  TUPLE_T *tuples;     // array of tuples within gene count range 
  int tuple_count;  // the number of valid tuples in tuples
  int total_counts; // the total number of counts added to the bin
};

/*
 * All of the bins in lookup will have their is_real value set
 * the same as is_real in this data type.
 */
typedef struct bins_t BINS_T;
struct bins_t {
  BIN_T **values; // values[gene-count] contains bin for gene-count
  int length; //allocated length of values array
  int total_tuples; //total number of tuples in bins
};


/************************************************************************
 * Bins, Bin and Tuple structure modification methods
 ***********************************************************************/

/*
 * Comparison function for two structures of TUPLE_T.
 * Compares based on the p-value of the tuple.
 */
inline int tuple_pvalue_compar(const void *v1, const void *v2) 
{
  TUPLE_T *t1 = (TUPLE_T*)v1;
  TUPLE_T *t2 = (TUPLE_T*)v2;
  if (t1->gomo_score < t2->gomo_score) {
    return -1;
  } else if (t1->gomo_score > t2->gomo_score) {
    return 1;
  } else {
    return 0;
  }
}

/*
 * Creates a BIN_T for holding tuples for GO terms that have been annotated
 * with a number of genes between the inclusive range of 'low_gene_count' and
 * 'high_gene_count'.
 *
 * 1. Allocate memory for the bin.
 * 2. Set all the values of the bin.
 * 3. Return the bin.
 */
BIN_T* bin_create(int low_gene_count, int high_gene_count) 
{
  assert(low_gene_count < high_gene_count);
  BIN_T *bin = mm_malloc(sizeof(BIN_T));
  bin->low_gene_count = low_gene_count;
  bin->high_gene_count = high_gene_count;
  bin->tuple_count = 0;
  bin->tuples = NULL;
  bin->total_counts = 0;
  return bin;
}

/*
 * Destroys a BIN_T created by bin_create.
 *
 * 1. If there are any tuples, free them.
 * 2. Free the bin.
 */
void bin_destroy(BIN_T *bin) 
{
  assert(bin != NULL);
  if (bin->tuple_count > 0) free(bin->tuples);
  free(bin);
}

/*
 * Adds a tuple created from the values 'go_term' and 'score'
 * to the specified bin with a counts value of zero.
 *
 * 1. Ensure there is enough space in the bin->tuples to add a tuple.
 * 2. Set the values of the tuple.
 * 3. Increment the tuple count.
 */
void bin_add_tuple(char *go_term, double score, BIN_T* bin) 
{
  if (bin->tuple_count % BIN_ALLOC_CHUNK == 0) {
    int new_size = bin->tuple_count + BIN_ALLOC_CHUNK;
    Resize(bin->tuples, new_size, TUPLE_T);
  }
  TUPLE_T* tuple = (bin->tuples)+(bin->tuple_count);
  tuple->go_id = go_term;
  tuple->gomo_score = score;
  tuple->counts = 0;
  bin->tuple_count += 1;
}

/*
 * Compares a key with an entry. The key is expected to be a pointer
 * to a double representing a p-value. The entry is expected to be a
 * pointer to a TUPLE_T which also contains a p-value. The two p-values
 * are compared and the function returns -1, 0 or 1 respectively if the
 * key's p-value is smaller, equal to, or greater, than the entry's p-value.
 */
inline int bin_bsearch_compar(const void *key, const void *entry) {
  double score = *((double*)key);
  TUPLE_T *tuple = (TUPLE_T*)entry;
  if (score < tuple->gomo_score) {
    return -1;
  } else if (score == tuple->gomo_score) {
    return 0;
  }
  return 1;
}

/*
 * Adds a count for a score generated by the null model. The counts allow emperical calculation
 * of the probability of a score occuring by random chance (p-value).
 *
 * 1. binary search the tuples to find a tuple that has the same score or a score just larger.
 * 2. if such a tuple exists then increment its count.
 */
void bin_add_count(double score, BIN_T* bin)
{
  int pos = binary_search(&score, bin->tuples, bin->tuple_count, sizeof(TUPLE_T), bin_bsearch_compar);
  if (pos < 0) pos = -(pos + 1);
  if (pos < bin->tuple_count) {
    (bin->tuples+pos)->counts += 1;
  }
  bin->total_counts += 1;
}


/*
 * Sort the tuples in the bin by p-value ascending.
 */
void bin_sort(BIN_T *bin) 
{
  qsort(bin->tuples, bin->tuple_count, sizeof(TUPLE_T), tuple_pvalue_compar);
}

/*
 * Create a bins data structure.
 * 
 * 1. Allocate memory for the bins structure and for the pointers to the bin types.
 * 2. Create the bin(s) so that there is a bin starting at zero and at each division 
 *    so that the entire possible range of gene counts has a bin allocated.
 * 3. Return the bins structure.
 */
BINS_T* bins_create(int divisions_length, // the number of places to split 
    int *divisions                         // places to spit the bins
    ) 
{
  assert(divisions_length >= 0);
  assert(divisions_length == 0 || divisions != NULL);
  BINS_T *bins = mm_malloc(sizeof(BINS_T));
  bins->length = 1 + divisions_length; 
  bins->values = mm_malloc(sizeof(BIN_T*)*bins->length);
  //create bins upto but not including the divisions
  int i, low;
  for (low = 0, i = 0; i < divisions_length; ++i) {
    assert(low < divisions[i]);
    bins->values[i] = bin_create(low, divisions[i]-1);
    low = divisions[i];
  }
  //create final bin
  bins->values[i] = bin_create(low, INT_MAX);
  bins->total_tuples = 0;
  return bins;
}

/*
 * Destroy a bin structure created by bins_create.
 */
void bins_destroy(BINS_T *bins) 
{
  assert(bins != NULL);
  int i;
  for (i = 0; i < bins->length; i++) {
    bin_destroy(bins->values[i]);
  }
  free(bins->values);
  free(bins);
}

inline int bins_bsearch_compar(const void *key, const void *entry) {
  int gene_count = *((int*)key);
  BIN_T *bin = *(BIN_T**)entry;
  if (gene_count >= bin->low_gene_count) {
    if (gene_count <= bin->high_gene_count) {
      return 0;
    } else {
      return 1;
    }
  }
  return -1;
}

/*
 * Use to get a bin for a gene_count. If the bin
 * doesn't exist then it returns NULL.
 */
BIN_T* bins_get_bin(int gene_count, BINS_T *bins) 
{
  assert(bins != NULL);
  assert(gene_count >= 0);
  int index = binary_search(&gene_count, bins->values, bins->length, sizeof(BIN_T*), bins_bsearch_compar);
  if (index < 0) return NULL;
  return bins->values[index];
}

/*
 * Bins a tuple based on the number of genes that are associated with the 'go_term'.
 */
void bins_add_bin_tuple(char *go_term, double score, int gene_count, BINS_T *bins) 
{
  assert(go_term != NULL);
  assert(bins != NULL);
  assert(gene_count >= 0); 
  bin_add_tuple(go_term, score, bins_get_bin(gene_count, bins));
  bins->total_tuples += 1;
}

/*
 * Bins a tuple based on the number of genes that are associated with the 'go_term'.
 */
void bins_add_bin_count(double score, int gene_count, BINS_T *bins) 
{
  bin_add_count(score, bins_get_bin(gene_count, bins));
}

/*
 * Sorts all the bin structures held by bins.
 */
void bins_sort(BINS_T *bins) 
{
  assert(bins != NULL);
  int i;
  for (i = 0; i < bins->length; ++i) {
    bin_sort(bins->values[i]);
  }
}

/*****************************************************************************
 * Methods relating to the SHUFFLER_T 
 *
 ****************************************************************************/
/*
 * shuffler_create
 * Creates a shuffler. The shuffling needs to be defined before this is used.
 */
SHUFFLER_T *shuffler_create(int gomap_seq_count, int total_seq_count) 
{
  assert(total_seq_count >= gomap_seq_count);
  SHUFFLER_T *shuffler = (SHUFFLER_T*)mm_malloc(sizeof(SHUFFLER_T));
  shuffler->map = (int*)mm_malloc(sizeof(int)*total_seq_count);
  shuffler->map_length = total_seq_count;
  shuffler->rank = (int*)mm_malloc(sizeof(int)*gomap_seq_count);
  shuffler->rank_length = gomap_seq_count;
  return shuffler;
}

/*
 * shuffler_destroy
 * Destroys a shuffler.
 */
void shuffler_destroy(SHUFFLER_T *shuffler) 
{
  free(shuffler->map);
  free(shuffler->rank);
}

/*
 * shuffler_identity
 * Sets the shuffler to the identity mapping
 */
void shuffler_identity(SHUFFLER_T *shuffler) 
{
  int i;
  for (i = 0; i < shuffler->map_length; ++i) {
    shuffler->map[i] = i;
  }
}

/*
 * bounded_random
 * Returns a random number between 0 (inclusive) and max (non inclusive).
 * Avoids bias by rerolling when the random source generates a number
 * outside of the range that can be evenly divided up amongst the options.
 * This has a non-deterministic run time, but nothing is perfect...
 */
inline int bounded_random(int max) 
{
  assert(max < RAND_MAX);
  int bound = RAND_MAX;
  int extra = RAND_MAX % max;
  if (extra != max - 1) {
    bound = RAND_MAX - extra - 1;
  }
  int rnd;
  do {
    rnd = rand();
  } while (rnd > bound);
  return rnd % max;
}

/*
 * shuffler_shuffle
 * Shuffles the map. Assumes that all map entries
 * are unique from a previous call to shuffler_identity
 */
void shuffler_shuffle(SHUFFLER_T *shuffler) 
{
  int i, s, *list, length;
  list = shuffler->map;
  length = shuffler->map_length;
  for (i = 0; i < length; ++i) {
    s = bounded_random(length - i) + i;
    int tmp = list[i];
    list[i] = list[s];
    list[s] = tmp;
  }
}

/*************************************************************************
 * gomo_result_cmp()
 *
 * Compares two gomo results with respect to its scores
 *************************************************************************/
inline int gomo_result_cmp(const void *v1, const void *v2) 
{
  assert(v1 != NULL);
  assert(v2 != NULL);
  GOMO_RESULT_T s1 = *(GOMO_RESULT_T *)v1;
  GOMO_RESULT_T s2 = *(GOMO_RESULT_T *)v2;
  if (s1.empirical_pvalue > s2.empirical_pvalue){
    return 1;
  } else if (s1.empirical_pvalue < s2.empirical_pvalue){
    return -1;
  } else {
    return 0;
  }
}

/*************************************************************************
 * start_xml_output()
 *
 * initiate the output
 *************************************************************************/
void start_xml_output(FILE *xml_output,
    int argc,
    char *argv[],
    char *go_map_file,
    ARRAYLST_T *ama_files,
    char *outdir,
    BOOLEAN_T clobber,
    BOOLEAN_T text_only,
    BOOLEAN_T use_e_values,
    double score_e_thresh,
    int min_gene_count,
    ARRAYLST_T *motifs,
    int shuffle_scores,
    double q_threshold
    )
{
  const char *archive_date = ARCHIVE_DATE;
  int a = strlen(archive_date);

  fprintf(xml_output,gomo_dts);
  fprintf(xml_output,
      "<gomo xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" version=\"%s\" release=\"%.*s\">\n",
      VERSION, a, archive_date);

  // create command line
  char *command = NULL;
  int lc = strlen(program_name);
  int alloc_size = lc + 1;
  int i;
  for (i=1;i<argc;++i){
    alloc_size += strlen(argv[i])+1;
  }
  command = mm_malloc(sizeof(char)*alloc_size);
  strncpy(command, program_name, lc);
  char* ptr;
  if (argc > 0) {
    for (i = 1; i < argc; ++i) {
      command[lc++] = ' ';
      ptr = argv[i];
      while (*ptr != '\0') {
        command[lc++] = *(ptr++);
      }
    }
  }
  command[lc] = '\0';
  // output program state 
  fprintf(xml_output,"\t<program name=\"%s\"\n"
      "\t\t\t cmd=\"%s\"\n"
      "\t\t\t outdir=\"%s\" clobber=\"%s\" text_only=\"%s\"\n"
      "\t\t\t use_e_values=\"%s\" score_e_thresh=\"%g\"\n"
      "\t\t\t min_gene_count=\"%d\"", 
      program_name, command, outdir, bool2txt(clobber), bool2txt(text_only), 
      bool2txt(use_e_values), score_e_thresh, min_gene_count);
  if (motifs != NULL) {
    fprintf(xml_output, " motifs=\"%s", (char*)arraylst_get(0, motifs));
    for (i = 1; i < arraylst_size(motifs); ++i) {
      fprintf(xml_output, " %s", (char*)arraylst_get(i, motifs));
    }
    fprintf(xml_output, "\"");
  }
  fprintf(xml_output, "\n\t\t\t shuffle_scores=\"%d\" q_threshold=\"%g\"\n\t\t\t>\n", shuffle_scores, q_threshold);
  fprintf(xml_output,"\t\t<gomapfile path=\"%s\" />\n", go_map_file);
  for (i = 0; i < arraylst_size(ama_files); ++i) {
    fprintf(xml_output,"\t\t<seqscorefile path=\"%s\" />\n", (char*)arraylst_get(i, ama_files));
  }
  fprintf(xml_output, "\t</program>\n");
  // clean up
  free(command);
}

inline double get_ama_score(const void *scores_ptr, int index) {
  AMA_LIST_T *scores = (AMA_LIST_T*)scores_ptr;
  return scores->list[index].score; 
}
inline BOOLEAN_T get_false(const void *ignore1, int ignore2) {
  return FALSE; 
}

/*
 * ama_score_compar
 * Compare two AMA scores. If the first is smaller, the same,
 * or larger than it will return repectively -1, 0, 1.
 */
int ama_score_compar(const void *v1, const void *v2) {
  AMA_SCORE_T *s1 = (AMA_SCORE_T*)v1;
  AMA_SCORE_T *s2 = (AMA_SCORE_T*)v2;
  if (s1->score < s2->score) {
    return -1;
  } else if (s1->score > s2->score) {
    return 1;
  }
  return 0;
}

/*
 * create_ama_score_list
 * Makes a copy of only the sequence scores and the identifying
 * sequence number (the order in which the sequence was loaded from file).
 * Additionally sorts the sequence list by the score ascending.
 */
AMA_LIST_T* create_ama_score_list(
    PATTERN_T *pattern,     // pattern to copy data from 
    HASH_TABLE seqid_ht,    // hash table for looking up sequences IN OUT
    ARRAYLST_T *seqid_keys, // list of keys for seqid ht IN OUT
    BOOLEAN_T use_e_values, // choose between the score source
    double ama_e_threshold // if the score source is converted p-values
                            // then this is the threshold at which the e-values
                            // are set to the maximum (meaning no association)
    ) 
{
  int seq_count = get_pattern_num_scanned_sequences(pattern);
  assert(seq_count > 0);
  // some useful variables
  AMA_LIST_T *ama_scores;
  AMA_SCORE_T *input;
  SCANNED_SEQUENCE_T *sseq;
  HASH_TABLE_ENTRY *he;
  int i, seq_num, *seq_num_mem;
  char *key;
  //set the sequence number to the next one
  seq_num = arraylst_size(seqid_keys);
  //allocate memory for the ama scores 
  ama_scores = mm_malloc(sizeof(AMA_LIST_T));
  ama_scores->list = mm_malloc(sizeof(AMA_SCORE_T)*seq_count);
  ama_scores->length = seq_count;
  //loop over all sequence scores
  for (i = 0, input = ama_scores->list; i < seq_count; ++i) {
    sseq = get_pattern_scanned_sequences(pattern)[i];
    he =  hash_lookup_str(get_scanned_sequence_name(sseq), seqid_ht);
    if (he == NULL) { // unknown sequence
      //put the sequence in to the sequence ht
      copy_string(&key, get_scanned_sequence_name(sseq));
      seq_num_mem = (int*)mm_malloc(sizeof(int));
      *seq_num_mem = seq_num;
      hash_insert_str_value(key, seq_num_mem, seqid_ht);
      arraylst_add(key, seqid_keys);
      // store the sequence loading number
      input->index = seq_num++;
    } else {
      // store the sequence loading number, rather than the sequence name as this speeds up lookup
      input->index = *((int*)hash_get_entry_value(he));
    }
    // select between copying e-values (calculated from pvalues) and sequence scores
    if (use_e_values) {
      if (!has_scanned_sequence_pvalue(sseq)) 
        die("The sequence score file does not provide a "
            "p-value for sequence %s (pattern %s). "
            "Use --gs to prompt GOMO working on "
            "gene scores rather than gene-score p-values.\n",
            get_scanned_sequence_name(sseq),
            get_pattern_accession(pattern));
      // combine the p-value with the sequence count to make an e-value
      input->score = get_scanned_sequence_pvalue(sseq)*seq_count;
      // if the e-value is above the specified threshold then consider the
      // association with the motif to be by random chance and score it accordingly
      if (ama_e_threshold > 0 && input->score > ama_e_threshold) {
        input->score = seq_count;
      }
    } else {
      if (!has_scanned_sequence_score(sseq)) 
        die("The sequence score file does not provide a "
            "score for sequence %s (pattern %s). "
            "Remove --gs to prompt GOMO working on "
            "gene-score p-values rather than gene scores.\n",
            get_scanned_sequence_name(sseq),
            get_pattern_accession(pattern));
      input->score = get_scanned_sequence_score(sseq);
    }
    ++input;
  }
  // sort the inputs in order of score ascending
  qsort(ama_scores->list, ama_scores->length, sizeof(AMA_SCORE_T), ama_score_compar);
  return ama_scores;
}

/*
 * destroy_ama_score_list
 * Frees all memory associated with the passed AMA_LIST_T.
 */
void destroy_ama_score_list(AMA_LIST_T *ama_scores) {
  free(ama_scores->list);
  free(ama_scores);
}


/******************************************************************************
 * load_unique_ids()
 *
 * Reads in the go map for a gomo run
 * Creates a list of unique go ids and a hash table with the go id as key
 * and the index and description in a goinfo structure as value. Also creates
 * a sequence hash table with the sequence id as key and its numbering as value.
 *****************************************************************************/
void load_unique_ids(
    char* go_map, // go map to scan IN
    ARRAYLST_T** goinfo_keys, // array list holding the set of unique go terms OUT
    HASH_TABLE* goinfo_ht, // hash table holding the description and index of go term OUT
    ARRAYLST_T** seq_keys, // arrays list holding the keys for seq_ht OUT
    HASH_TABLE* seq_ht, // hash table hold the sequences of go terms OUT
    int min_gene_count  // minimum gene count to be considered for loading
    ) 
{
  assert(go_map != NULL);
  assert(goinfo_keys != NULL);
  assert(goinfo_ht != NULL);
  assert(seq_keys != NULL);
  assert(seq_ht != NULL);
  assert(min_gene_count > 0);

  FILE *inputFilePtr;             /* Pointer to input file */
  STRING_LIST_T **content = NULL; /* content of input file */
  int i, j, length = 0, gene_count, gonum = 0, seqnum = 0, *num;
  char *goid, *godesc, *seqid;
  BOOLEAN_T success;
  GOINFO_T *goinfo;
  
  if (status) fprintf(stderr, "Loading unique ids from \"%s\".\n", go_map);

  inputFilePtr = fopen(go_map, "r");   /* Open in TEXT mode */
  read_csv(inputFilePtr,'#','\t',&content,&length);     /* Get csv content*/
  fclose(inputFilePtr);

  //estimate a size, double should be more than enough
  *goinfo_ht = hash_create(length*2);
  *goinfo_keys = arraylst_create_sized(length);
  *seq_ht = hash_create(length*2); //a better way of estimating this would be nice
  *seq_keys = arraylst_create();
  num = (int*)mm_malloc(sizeof(int));
  *num = seqnum;
  for (i = 0; i < length; ++i) { 
    gene_count = get_num_strings(content[i]) - 2;
    if (gene_count >= min_gene_count) {
      goid = get_nth_string(0, content[i]);
      godesc = get_nth_string(1, content[i]);
      goinfo = mm_malloc(sizeof(GOINFO_T));
      goinfo->index = gonum++;
      copy_string(&(goinfo->desc), godesc);
      if (!hash_insert_str_value(goid, goinfo, *goinfo_ht)) {
        die("GO term '%s' appears more than once!", goid);
      }
      // get the key from the goinfo_ht and store it in our list
      arraylst_add(hash_get_entry_key(hash_lookup_str(goid, *goinfo_ht)), *goinfo_keys);
      // put each new sequence in the sequence hash table
      for (j = 2; j < get_num_strings(content[i]); ++j) {
        seqid = get_nth_string(j, content[i]);
        success = hash_insert_str_value(seqid, num, *seq_ht); // num is always allocated ahead of time
        if (success) {
          // get the key from seq_ht and store it in our list
          arraylst_add(hash_get_entry_key(hash_lookup_str(seqid, *seq_ht)), *seq_keys);
          num = (int*)mm_malloc(sizeof(int));
          *num = ++seqnum;
        }
      }
    }
  }
  // deallocate excess memory
  arraylst_fit(*goinfo_keys);
  arraylst_fit(*seq_keys);
  
  //clean up
  free(num);//always allocated ahead of time, so there will be one left over to free
  for (i=0; i < length; i++) {
    // free content
    free_string_list(content[i]);
  }
  myfree(content);

  if (verbosity >= NORMAL_VERBOSE) {
    fprintf(stderr, "Loaded %d unique GO terms (IDs and descriptions) with %d or more annotations and %d annotating sequence IDs.\n", 
        arraylst_size(*goinfo_keys), min_gene_count, arraylst_size(*seq_keys));
  }
}

/******************************************************************************
 * load_go_map()
 *
 * Reads in go term to sequence mapping from file
 *****************************************************************************/
SEQ_LIST_T* load_go_map(
    char* filename,         //the file to read IN 
    HASH_TABLE goinfo_ht,   //the hash table of go terms IN 
    HASH_TABLE sequence_ht, //the hash table of sequence ids IN
    int min_gene_count
  ) 
{
  assert(filename != NULL);
  assert(goinfo_ht != NULL);
  assert(sequence_ht != NULL);
  // some useful variables
  FILE *inputFilePtr;             /* Pointer to input file */
  STRING_LIST_T **lines = NULL;
  int i, j, line_count = 0, gene_count, goids_loaded = 0;
  HASH_TABLE_ENTRY* he = NULL;

  if (status) fprintf(stderr, "Loading GO map from \"%s\".\n", filename);
  
  // load in the file
  inputFilePtr = fopen(filename, "r");    /* Open in TEXT mode */
  read_csv(inputFilePtr, '#','\t', &lines, &line_count); /* Get csv lines*/
  fclose(inputFilePtr);
  if (verbosity >= HIGH_VERBOSE) {
    fprintf(stderr, "Read in %d lines.\n", line_count);
  }
  // allocate mappings
  SEQ_LIST_T *go_map = (SEQ_LIST_T*)mm_calloc(get_num_entries(goinfo_ht), sizeof(SEQ_LIST_T));
  // now process the file
  for (i = 0; i < line_count; ++i) { /* read in each and every line */
    gene_count = get_num_strings(lines[i]) -2;
    if (gene_count <= 0) {
      if (verbosity >= NORMAL_VERBOSE) {
        fprintf(stderr, "Line %d is empty or has GO term without assigned sequences. Skip this line!\n", i+1);
      }
    } else if (gene_count >= min_gene_count) {
      SEQ_LIST_T * sequences;
      int size_to_allocate = get_num_strings(lines[i]) - 2;
      // lookup the go term
      char* go_term = get_nth_string(0,lines[i]);
      he = hash_lookup_str(go_term, goinfo_ht);
      if (he == NULL) { //we've never seen it before!
        die("New go term '%s' encountered in map file '%s'.\n", go_term, filename);
      }
      GOINFO_T *goinfo = (GOINFO_T*)hash_get_entry_value(he);
      sequences = go_map+(goinfo->index);
      if (sequences->seqnums == NULL) {
        sequences->length = size_to_allocate;
        sequences->seqnums = (int*)mm_malloc(sizeof(int)*size_to_allocate);
        ++goids_loaded;
      } else {
        die("GO term %s appears more than once.\n", go_term);
      }
      // link sequence ids to go term so there only
      // exists one allocation of each sequence id
      for (j = 2; j < get_num_strings(lines[i]); ++j) {
        char* seq_id = get_nth_string(j,lines[i]);
        he = hash_lookup_str(seq_id, sequence_ht);
        if (he == NULL) {
          die("New sequence '%s' encountered in shuffled map!\n", seq_id);
        }
        sequences->seqnums[j - 2] = *((int*)hash_get_entry_value(he));
      }
    }
  }
  if (goids_loaded != get_num_entries(goinfo_ht)) {
    die("Less go ids in go map than expected!\n");
  }
  // free lines
  for (i = 0; i < line_count; ++i) {
    free_string_list(lines[i]);
  }
  myfree(lines);
  return go_map;
}

/******************************************************************************
 * load_ama_scores
 * Loads CisML files containing average motif afinity scores for sequences and
 * extracts the sequence scores, the type of which is dependent on the use_e_values
 * parameter. Creates a hash table with the motif as key and the sequence scores 
 * for each species as value. If selected_motifs is not null then it will only store
 * those motifs.
 *****************************************************************************/
void load_ama_scores(
    ARRAYLST_T *ama_files,      // the list of score files to source ama scores
    HASH_TABLE seqid_ht,        // the hash table of previously loaded sequence ids IN OUT
    ARRAYLST_T *seqid_keys,     // the keys to the sequence ids hash table IN OUT
    BOOLEAN_T use_e_values,     // the type of score to extract from the CisML file
    double ama_e_threshold,     // the threshold for considering random for a e-value
    ARRAYLST_T *selected_motifs,// the motifs to limit analysis to, or null
    HASH_TABLE *motif_ht,       // a constructed hash table containing the motifs->scores OUT
    LINKLST_T **motif_ll        // the motif keys in the motifs_ht OUT
  ) 
{
  assert(ama_files != NULL);
  assert(arraylst_size(ama_files) > 0);
  assert(seqid_ht != NULL);
  assert(seqid_keys != NULL);
  // some useful variables
  int i, j;
  PATTERN_T *pattern;
  char* accession;
  AMA_LIST_T **species_scores;
  // load shared motifs 
  *motif_ll = linklst_create();
  size_t input_list_memsize = sizeof(AMA_LIST_T*)*arraylst_size(ama_files);
  for (i = 0; i < arraylst_size(ama_files); ++i) {
    char *ama_file = (char*)arraylst_get(i, ama_files);
    if (status) fprintf(stderr, "Loading scores from \"%s\".\n", ama_file);
    CISML_T *file = read_cisml(ama_file);
    if (file == NULL) {
      // Wasn't cisml XML
      die("The scoring file needs to be in CisML format: %s",
          ama_file);
    }
    if (verbosity >= HIGHER_VERBOSE) {
      fprintf(stderr, "Cisml file '%s' for species %d contains %d motifs with sequence scores.\n", 
          ama_file, i, get_cisml_num_patterns(file));
    }
    if (i == 0) {
      // after we know the upper bound on the motif count, create the hash table
      // the hash table will have the motif as key and per-species lists of the motif's ama scores as values
      *motif_ht = hash_create(get_cisml_num_patterns(file)*2);
      for (j = 0; j < get_cisml_num_patterns(file); ++j) {
        pattern = get_cisml_patterns(file)[j];  
        accession = get_pattern_accession(pattern); // motif identifier
        if (selected_motifs != NULL) {
          // if this motif hasn't been selected than skip it
          if (arraylst_bsearch(accession, arraylst_compar_txt, selected_motifs) < 0) continue;
        }
        // allocate memory to for the per-species lists of ama scores for this motif
        species_scores = (AMA_LIST_T**)mm_malloc(input_list_memsize);
        memset(species_scores, 0, input_list_memsize);
        // make a copy of everything useful in the pattern
        species_scores[0] = create_ama_score_list(pattern, seqid_ht, seqid_keys, use_e_values, ama_e_threshold);
        // associate the per-species sequence scores with the motif
        if (!hash_insert_str_value(accession, species_scores, *motif_ht)) {
          // the motif already has scores, this is strange so we should warn the user
          fprintf(stderr, "Pattern %s appears more than once in cisml file '%s'."
              " Only the first occurance will be tested.\n", accession,
              ama_file);
          // clean up
          destroy_ama_score_list(species_scores[0]);
          free(species_scores);
        } else {
          // the motif is definately new, so we add the key to the keylist
          linklst_add(hash_get_entry_key(hash_lookup_str(accession, *motif_ht)), *motif_ll);
        }
      }
    } else {
      HASH_TABLE_ENTRY *he;
      for (j = 0; j < get_cisml_num_patterns(file); ++j) {
        pattern = get_cisml_patterns(file)[j];  
        accession = get_pattern_accession(pattern);
        // check the motif has been in previous cisml files, otherwise skip
        if ((he = hash_lookup_str(accession, *motif_ht)) != NULL) {
          // we've seen this motif before, which is good
          species_scores = hash_get_entry_value(he); 
          if (species_scores[i-1] == NULL) {
            // the previous cisml file has not seen this motif though so we must remove it
            // as we can only work on motifs that all species have data for 
            hash_remove_str(accession, *motif_ht);
            // clean up
            int k = 0;
            while (species_scores[k] != NULL) {
              destroy_ama_score_list(species_scores[k]);
              ++k;
            }
            free(species_scores);
            continue;
          }
          if (species_scores[i] != NULL) {
            // the motif already has scores, this is strange so we should warn the user
            fprintf(stderr, "Pattern %s appears more than once in cisml file '%s'."
                " Only the first occurance will be tested.\n", accession,
                ama_file);
          } else {
            // make a copy of everything useful in the pattern
            species_scores[i] = create_ama_score_list(pattern, seqid_ht, seqid_keys, use_e_values, ama_e_threshold);
          }
        }
      }
    }
    // clean up
    free_cisml(file);
  }
  // remove any items in motif_ll that are no-longer in motif_ht
  LINK_T *link = linklst_first(*motif_ll);
  while (link != NULL) {
    if (hash_lookup_str(link_item(link), *motif_ht) == NULL) {
      LINK_T *next = link_next(link);
      linklst_remove(link, *motif_ll);
      link = next;
    } else {
      link = link_next(link);
    }
  }
  if (verbosity >= NORMAL_VERBOSE) {
    fprintf(stderr, "Loaded %d motifs which have sequence scores for all species.\n", 
        linklst_size(*motif_ll));
  }
}


/*************************************************************************
 * calculate_ranksum_pvalues
 * Calculates a p-value for each GO-term using the Mann-Whitney U test.
 *
 * For each GO-term the sequences are classified into two groupings, those
 * that are annotated with the GO-term and those that are not. The ranksum
 * test (aka Mann-Whitney U test) is run to determine if the scores from the 
 * two classifications are from the same distribution giving a p-value.
 *
 *************************************************************************/
double *calculate_ranksum_pvalues(
    ARRAYLST_T *all_go_terms,
    SEQ_LIST_T *go_map,
    AMA_LIST_T *seq_scores,
    SHUFFLER_T *shuffler,
    BOOLEAN_T null_score
  ) 
{
  int i, j;
  int seq_count = seq_scores->length;

  // create mapping of sequence num to its index in the ordered list of scores
  for (i = 0; i < shuffler->rank_length; ++i) {
    shuffler->rank[i] = -1;
  }
  for (i = 0; i < seq_scores->length; ++i) {
    int index = seq_scores->list[i].index;
    index = shuffler->map[index];
    if (index < shuffler->rank_length) {
      shuffler->rank[index] = i;
    }
  }

  // class of each sequence
  // this is used by the Mann-Whitney U test to define the two groups
  // that it is trying to determine if they are from the same distribution
  // in the case of this method the grouping will define which sequences
  // are linked to the current go term under test
  BOOLEAN_T *group = (BOOLEAN_T*)mm_malloc(sizeof(BOOLEAN_T)*seq_count);

  // Copy the scores and the uninitialized classes (group) to
  // a special object for ranksum routines.  Calculate the
  // (adjusted) rank of each score.
  // The group of each score will be reassigned differently for
  // each GO term.
  // By adjusted rank this means that all consecutive equal scores
  // will have their new ranks made the average of their ranks.
  // Note that the copying of the group is completely pointless
  // here because the group variable is not yet set but there
  // is no way around this because the grouping has to be redone
  // for each go-term.
  RSD_T *ranksum_dataset = get_ranksum_dataset2(seq_scores, get_ama_score, group, get_false, seq_count);

  int skipped = 0;
  double *go_scores = (double*)mm_malloc(sizeof(double)*arraylst_size(all_go_terms));
  // for each go term mark all the mapped sequences as being in class a 
  for (i=0; i < arraylst_size(all_go_terms); ++i) {
    char *go_term = arraylst_get(i, all_go_terms);
    // number of elements in class a
    int class_a = 0;
    // define all entries as not belonging to the GO-term by default
    for (j = 0; j < seq_count; ++j) {
      group[j] = FALSE;
    }
    // look up the sequence ids associated with the go id 
    SEQ_LIST_T *seqids = go_map+i;

    // for each sequence id look up its rank
    for (j = 0; j < seqids->length; ++j) {
      // lookup the rank of the sequence id and
      // set the the sequence id to be a member
      // of group a
      int index = shuffler->rank[seqids->seqnums[j]];
      if (index != -1) {
        group[index] = TRUE;
        ++class_a;
      }
    }
    // can not calculate if all are in one classification 
    if (class_a < seq_count && class_a > 0) {
      set_ranksum_groups(ranksum_dataset, group);
      RSR_T *ranksum_result = run_ranksum_test_on_ordered_dataset(ranksum_dataset);
      go_scores[i] = RSR_get_p_left(ranksum_result);
      destroy_rsr(ranksum_result);
    } else {
      //no data
      if (null_score) {
        // as we're doing lots of samples a random function should average out
        go_scores[i] = drand48();
      } else {
        // only a few samples so can't use rand.
        go_scores[i] = 0.5; 
      }
      ++skipped;
      if (verbosity >= HIGHER_VERBOSE) {
        fprintf(stderr, "GO term %s skipped as either all sequences were in its class or none were (%d / %d).\n", go_term, class_a, seq_count);
      }
    }
  }
  if (verbosity >= HIGH_VERBOSE) {
    if (skipped > 0)
      fprintf(stderr,"%d GO terms have assigned genes which lack a sequence entry"
          " or for other reasons have no information to work with. These are "
          "skipped!\n",skipped);
  }
  
  // clean up
  free(group);
  destroy_rsd(ranksum_dataset);

  return go_scores;
}

/*************************************************************************
 * calculate_scores_and_bin
 * Compute the GOMO multiple-species score for the motif for each GO-term.
 * Those scores are saved in "bins" indexed by number of genes annotated
 * to a GO-term.
 *
 *   1) Loops over species calling calculate_ranksum_pvalues, which 
 *      returns an array with all the single-species GOMO scores.
 *   2) Computes the multiple-species GOMO score for each GOterm by
 *	    computing the geometric mean over species.
 *   3) If the data is not from the null model then tuples are added
 *      recording the goid and score, otherwise counts are added to the
 *      tuples 
 ************************************************************************/
void calculate_scores_and_bin(
    ARRAYLST_T *all_go_terms, 
    SEQ_LIST_T* go_map, 
    AMA_LIST_T **motif_scores, 
    int species_count, 
    BOOLEAN_T is_nmod,
    SHUFFLER_T *shuffler,
    BINS_T *bins
  ) 
{
  int i, j;
  // allocate memory for storing a p value for every go term per species
  double ** go_scores =  (double**)mm_malloc(sizeof(double*)*species_count);
  // loop on species
  for (i = 0; i < species_count; ++i) {
    go_scores[i] = calculate_ranksum_pvalues(all_go_terms, go_map, motif_scores[i], shuffler, is_nmod);
  }
  // loop on go ids
  for (i = 0; i < arraylst_size(all_go_terms); ++i) {
    char *go_id = arraylst_get(i, all_go_terms);
    //calculate geometric mean
    double score = 1.0;
    for (j = 0; j < species_count && score > 0; ++j) {
      //Note: if the p value wasn't calculated it will be negative
      score *= go_scores[j][i];
    }
    score = pow(score, 1.0 / species_count);// take nth root
    //add tuple
    int gene_count = go_map[i].length;
    if (is_nmod) {
      bins_add_bin_count(score, gene_count, bins);
    } else {
      bins_add_bin_tuple(go_id, score, gene_count, bins);
    }
  }
  // clean up
  for (i = 0; i < species_count; ++i) {
    free(go_scores[i]);
  }
  free(go_scores);
}

/**************************************************************
 * calculate_qvalues_and_output
 * Current motif p-values are computed empirically for each gene-count bin,
 * and outputs all GOterm GOMO scores (etc) for the motif. 
 * 
 * The empirical p-values for all GO-terms in a bin are estimated
 * using the counts of null model p-values that scored lower or equal. 
 * Then the p-values of all bins are input to the compute_q_value() function.
 * The E-values are computed by multiplying the empirical p-value by the
 * number of p-values.
 *************************************************************/
void calculate_qvalues_and_output(
    char *motif, 
    BINS_T *bins, 
    HASH_TABLE goinfo_ht, 
    FILE *xml_output,
    BOOLEAN_T text_only,
    double q_threshold,
    int species_count
    ) 
{
  assert(motif != NULL);
  assert(bins != NULL);
  
  GOMO_RESULT_T *results = (GOMO_RESULT_T*)mm_malloc(sizeof(GOMO_RESULT_T)*bins->total_tuples);

  int i, j, k, num_dupes, running_total; 
  GOMO_RESULT_T *result = results;
  for (i = 0; i < bins->length; ++i) {
    BIN_T *bin = bins->values[i];
    running_total = 0;
    for (j = 0; j < bin->tuple_count; j += num_dupes) {
      num_dupes = 0;
      // if there is only one species the q-value can be calculated from the rst p-values
      /*if (species_count == 1) { //FIXME disabled while a decision is made on the validity of this optimisation
        ++num_dupes;
        result->go_id = bin->tuples[j].go_id;
        result->gomo_score = bin->tuples[j].gomo_score;
        result->empirical_pvalue = bin->tuples[j].gomo_score; // not really an empirical pvalue
        ++result;
      } else {*/
        // check if any of the next tuples have duplicate scores
        // as we will need to include them in this step
        do {
          running_total += bin->tuples[j+num_dupes].counts;
          ++num_dupes;
        } while (j+num_dupes < bin->tuple_count && bin->tuples[j].gomo_score == bin->tuples[j+num_dupes].gomo_score);
        double empirical_pvalue  = (double)(max(running_total, 1)) / (double) (bin->total_counts + 1);
        for (k = 0; k < num_dupes; ++k) { 
          result->go_id = bin->tuples[j+k].go_id;
          result->gomo_score = bin->tuples[j].gomo_score;
          result->empirical_pvalue = empirical_pvalue; 
          ++result;
        }
      //} //FIXME disabled while a decision is made on the validity of this optimisation
    }
  }
  qsort(results, bins->total_tuples, sizeof(GOMO_RESULT_T), gomo_result_cmp); 
  //copy to ARRAY_T for processing by compute_qvalues
  ARRAY_T *score_array = allocate_array(bins->total_tuples);
  for (i = 0; i < bins->total_tuples; ++i) 
    set_array_item(i, results[i].empirical_pvalue, score_array);
  // compute the qvalues
  compute_qvalues(FALSE, // Don't stop with FDR.
      TRUE,  // Do estimate pi-zero.
      NULL,  // Don't store pi-zero in a file.
      NUM_BOOTSTRAPS,
      NUM_BOOTSTRAP_SAMPLES,
      NUM_LAMBDA,
      MAX_LAMBDA,
      bins->total_tuples,
      score_array);
  //output motif's results to xml
  if (!text_only) fprintf(xml_output,"\t<motif id=\"%s\">\n", motif);
  for (i = 0; i < bins->total_tuples; ++i) {
    char *go_id = results[i].go_id;
    double score = results[i].gomo_score;
    double pvalue = results[i].empirical_pvalue;
    double qvalue = get_array_item(i, score_array);

    if (qvalue > q_threshold) break;

    if (text_only) {
      fprintf(stdout, "%s\t%s\t%1.3e\t%1.3e\t%1.3e\n",
          motif, go_id, score, pvalue, qvalue);
    } else {
      char *go_desc = ((GOINFO_T*)hash_get_entry_value(hash_lookup_str(go_id, goinfo_ht)))->desc;
      fprintf(xml_output,"\t\t<goterm id=\"%s\" score=\"%1.3e\" pvalue=\"%1.3e\" qvalue=\"%1.3e\"\n\t\t\t\tdescription=\"%s\"/>\n",
          go_id, score, pvalue, qvalue, go_desc);
    }
  }
  if (!text_only) fprintf(xml_output,"\t</motif>\n");
  //clean up
  free_array(score_array);
  free(results);
}

/*
 * Attempt to parse an integer from the string.
 * If the conversion fails then display the passed error message and exit.
 */
int parse_integer(char *str, char *no_conversion_msg, char *partial_conversion_msg, char *underflow_msg, char *overflow_msg) {
  char *end;
  long l;
  errno = 0;
  l = strtol(str, &end, 10);
  if (end == str) {
    if (no_conversion_msg) die(no_conversion_msg, str);
  } else if (*end != '\0') {
    if (partial_conversion_msg) die(partial_conversion_msg, end);
  }
  if (errno == ERANGE || l < INT_MIN || l > INT_MAX) {
    if (l < INT_MIN) { // underflow
      if (underflow_msg) die(underflow_msg, str);
    } else {
      if (overflow_msg) die(overflow_msg, str);
    }
  }
  return (int)l;
}

/*
 * Attempt to parse a double from the string.
 * If the conversion fails then display the passed error message and exit.
 */
double parse_double(char *str, char *no_conversion_msg, char *partial_conversion_msg, char *underflow_msg, char *overflow_msg) {
  char *end;
  double d;
  errno = 0;
  d = strtod(str, &end);
  if (end == str) { // no conversion
    if (no_conversion_msg) die(no_conversion_msg, str);
  } else if (*end != '\0') {
    if (partial_conversion_msg) die(partial_conversion_msg, end);
  }
  if (errno == ERANGE) {
    if (d == 0) { // underflow
      if (underflow_msg) die(underflow_msg, str);
    } else {
      if (overflow_msg) die(overflow_msg, str);
    }
  }
  return d;
}

/*****************************************************************************
 * process_cmd_line
 * Extract the required information out of the command line.
 *****************************************************************************/
void process_cmd_line(
    int argc, 
    char *argv[], 
    char **output_dirname, 
    BOOLEAN_T *clobber, 
    BOOLEAN_T *text_only,
    ARRAYLST_T **selected_motifs, 
    BOOLEAN_T *use_E_values, 
    double *score_E_thresh,
    int *shuffle_scores,
    int *min_gene_count,
    double *q_threshold,
    char ** go_map_file,
    ARRAYLST_T **ama_files
  ) 
{
  // set defaults
  *output_dirname = default_output_dirname;
  *clobber = TRUE;
  *text_only = FALSE;
  *selected_motifs = NULL;
  *use_E_values = TRUE;
  *score_E_thresh = 0.0;
  *shuffle_scores = DEFAULT_SHUFFLES;
  *min_gene_count = DEFAULT_MIN_GENE_COUNT;
  *q_threshold = DEFAULT_Q_VALUE_THRESHOLD;

  /**********************************************
   * COMMAND LINE PROCESSING
   **********************************************/

  const int num_options = 11;
  cmdoption const motif_scan_options[] = {
    {"o", REQUIRED_VALUE},
    {"oc", REQUIRED_VALUE},
    {"text", NO_VALUE},
    {"shuffle_scores", REQUIRED_VALUE},
    {"motif",REQUIRED_VALUE},
    {"score_E_thresh", REQUIRED_VALUE},
    {"t", REQUIRED_VALUE},
    {"min_gene_count", REQUIRED_VALUE},
    {"gs", NO_VALUE},
    {"nostatus", NO_VALUE},
    {"verbosity", REQUIRED_VALUE}
  };

  int option_index = 0;

  // Define the usage message.
  char usage[] =
    "USAGE: gomo [options] <go-map file> <scoring file>+\n"
    "\n"
    "   Options:\n"
    "     --o <output dir>\t\t"         "Name of the directory for output. Will not replacing an existing directory.\n"
    "     --oc <output dir>\t\t"        "Name of the directory for output. Will replace an existing directory. (default='gomo_out')\n"
    "     --text\t\t\t"                 "FLAG Output in tab separated values format to standard out. Will not create a output directory or files.\n"
    "     --motif <id>\t\t"             "Limit results to specified motif. Option may be repeated with additional motifs. (default=all)\n"
    "     --shuffle_scores <n>\t"       "Generate empirical nulls by shuffling the sequence = score assignment. Repeat n times. (default=10)\n"
    "     --score_E_thresh <double>\t"  "The E-value threshold above which all scored sequences obtain the same rank. (default=no threshold)\n"
    "     --t <double>\t\t"             "The q-value threshold considered significant. (default = 0.05, t >= 1.0 shows all results)\n"
    "     --min_gene_count <n>\t"       "Only load go terms that are annotated with a minimum of n genes. (default = 1).\n"
    "     --gs\t\t\t"                   "FLAG Use gene scores rather than p-values for the calculations.\n"
    "     --nostatus\t\t\t"             "FLAG Prevent reporting of progress updates to stderr.\n"
    "     --verbosity [1|2|3|4]\t"      "1 = Quiet, 2 = Normal, 3 = High, 4 = Very High. (default 2)\n"
    "\n";

  // Parse the command line.
  if (simple_setopt(argc, argv, num_options, motif_scan_options) != NO_ERROR) {
    die("Error processing command line options: option name too long.\n");
  }

  while (TRUE) {
    int c = 0;
    char* option_name = NULL;
    char* option_value = NULL;
    const char * message = NULL;

    // Read the next option, and break if we're done.
    c = simple_getopt(&option_name, &option_value, &option_index);
    if (c == 0) {
      break;
    } else if (c < 0) {
      (void) simple_getopterror(&message);
      die("Error processing command line options (%s)\n", message);
    }
    //output directory if not already
    if (strcmp(option_name, "o") == 0) {
      *output_dirname = option_value;
      *clobber = FALSE;
    }
    //output to directory always
    else if (strcmp(option_name, "oc") == 0) {
      *output_dirname = option_value;
      *clobber = TRUE;
    }
    else if (strcmp(option_name, "text") == 0) {
      *text_only = TRUE;  
    }
    //only calculate for selected motifs
    else if (strcmp(option_name, "motif") == 0) {
      if (*selected_motifs == NULL) {
        *selected_motifs = arraylst_create();
      }
      arraylst_add(option_value, *selected_motifs);
    }
    //shuffle the ama sequence ids n times
    else if (strcmp(option_name, "shuffle_scores") == 0) {
      //test that the option is only digits
      int n = parse_integer(option_value, 
          "The option shuffle_scores was set to \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option shuffle_scores ended with \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option shuffle_scores was set to %s, it requires a "
              "number larger than zero.\n",
          "The option shuffle_scores was set to %s which "
              "could not be used because the value is too large to "
              "represent internally.\n");
      if (n < 1)
        die ("The option shuffle_scores was set to %d, it requires a "
            "number larger than zero.\n", n);
      *shuffle_scores = n;
    }
    //E-values (from ama) above the threshold are ranked the same
    else if (strcmp(option_name, "score_E_thresh") == 0) {
      double x = parse_double(option_value, 
          "The option score_E_thresh was set to \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option score_E_thresh ended with \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option score_E_thresh was set to %s, which "
              "could not be used because the value requires more precision "
              "then is possible to represent internally.\n",
          "The option score_E_thresh was set to %s which "
              "could not be used because the value is too large to "
              "represent internally.\n");
      if (x <= 0.0)
        die ("The option score-E-thresh was set to %f, it requires a "
            "number larger than zero.\n", x);
      *score_E_thresh = x;
    }
    //threshold of GOMO qvalues expected to report
    else if (strcmp(option_name, "t") == 0) {
      double x = parse_double(option_value, 
          "The option t was set to \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option t ended with \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option t was set to %s, which "
              "could not be used because the value requires more precision "
              "then is possible to represent internally.\n",
          "The option t was set to %s which "
              "could not be used because the value is too large to "
              "represent internally.\n");
      if (x < 0.0 || x > 1.0)
        die ("The option t was set to %f, it requires a "
            "number between 0.0 and 1.0.\n", x);
      *q_threshold = x;
    }
    //only calculate for go ids that are annotated with more genes than the minimum gene count 
    else if (strcmp(option_name, "min_gene_count") == 0) {
      int n = parse_integer(option_value, 
          "The option min_gene_count was set to \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option min_gene_count ended with \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option min_gene_count was set to %s, it requires a "
              "number larger than zero.\n",
          "The option min_gene_count was set to %s which "
              "could not be used because the value is too large to "
              "represent internally.\n");
      if (n < 1)
        die ("The option min_gene_count was set to %d, it requires a "
            "number larger than zero.\n", n);
      *min_gene_count = n;
    }
    //use gene scores instead of converting p-values into e-values
    else if (strcmp(option_name, "gs") == 0) {
      *use_E_values = FALSE;
    }
    //don't report any status information
    else if (strcmp(option_name, "nostatus") == 0) {
      status = FALSE;
    }
    //the level of output
    else if (strcmp(option_name, "verbosity") == 0) {
      int n = parse_integer(option_value, 
          "The option verbosity was set to \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option verbosity ended with \"%s\" which could "
              "not be interpreted as a number.\n",
          "The option verbosity was set to %s, it requires a "
              "number between 1 and 4 inclusive.\n",
          "The option verbosity was set to %s it requires a "
              "number between 1 and 4 inclusive.\n");
      if (n < 1 || n > 5)
        die ("The option verbosity was set to %d, it requires a "
            "number between 1 and 4 inclusive.\n", n);
      verbosity = n; 
    }
  }

  // if we had any motifs then sort them
  if (*selected_motifs != NULL) {
    arraylst_qsort(arraylst_compar_txt, *selected_motifs);
  }

  // Must have go-map directory and at least one scoring file
  if (argc < option_index + 2) {
    fprintf(stderr, usage);
    exit(EXIT_FAILURE);
  }

  *go_map_file = argv[option_index++];

  if (!file_exists(*go_map_file)) {
    die("GO map file \"%s\" does not exist.\n", *go_map_file);
  }

  *ama_files = arraylst_create();
  while (option_index < argc) { 
    char *score_file = argv[option_index++];
    if (!file_exists(score_file)) {
      die("Scoring file \"%s\" does not exist.\n", score_file);
    }
    arraylst_add(score_file,*ama_files);
  }
}


/*************************************************************************
 * Entry point for gomo
 *************************************************************************/
int main(int argc, char *argv[]) 
{
  int i, j, shuffle_scores, min_gene_count, gomap_seq_count, species_count, status_update_iter;
  double score_E_thresh, q_threshold;
  char *output_dirname, *go_map_file, *xml_path;
  BOOLEAN_T clobber, text_only, use_Evalues;
  ARRAYLST_T *selected_motifs, *ama_files, *goinfo_keys, *seqid_keys;
  HASH_TABLE seqid_ht, goinfo_ht, motif_ht;
  HASH_TABLE_ENTRY *he;
  SEQ_LIST_T *gomap;
  LINKLST_T *motif_ll;
  LINK_T *link;
  SHUFFLER_T *shuffler;
  FILE *xml_output;
  clock_t c0, c1;

  process_cmd_line(
      argc, 
      argv,
      &output_dirname,
      &clobber,
      &text_only,
      &selected_motifs,
      &use_Evalues,
      &score_E_thresh,
      &shuffle_scores,
      &min_gene_count,
      &q_threshold,
      &go_map_file,
      &ama_files
    );

  if (status) fprintf(stderr, "Starting\n");

  species_count = arraylst_size(ama_files);

  // how often to output a status message when shuffling
  status_update_iter = min(50, max(10, shuffle_scores / 10));

  set_alphabet(verbosity, "ACGT");

  // measure time
  c0 = clock();

  //set seeds on pseudo random number generators
  srand(0);
  srand48(0);

  // load all ids for go terms and sequences, additionally load descriptions for go terms 
  load_unique_ids(go_map_file, &goinfo_keys, &goinfo_ht, &seqid_keys, &seqid_ht, min_gene_count);
  gomap_seq_count = arraylst_size(seqid_keys);

  // load the go map
  gomap = load_go_map(go_map_file, goinfo_ht, seqid_ht, min_gene_count);

  // load ama scores
  load_ama_scores(ama_files, seqid_ht, seqid_keys, use_Evalues, score_E_thresh, selected_motifs, &motif_ht, &motif_ll);

  // create shuffler
  shuffler = shuffler_create(gomap_seq_count, arraylst_size(seqid_keys));
  
  if (!text_only) {
    // configure output
    if (create_output_directory(output_dirname, clobber, (verbosity >= NORMAL_VERBOSE))) {
      // Failed to create output directory.
      exit(1);
    }
    // Create the name of the text output file
    // "<dir>/XML_FILENAME" and open it for writing
    xml_path = make_path_to_file(output_dirname, XML_FILENAME);
    xml_output = fopen(xml_path, "w");

    //start output
    start_xml_output(xml_output, argc, argv, go_map_file, ama_files, 
        output_dirname, clobber, text_only, use_Evalues, score_E_thresh, 
        min_gene_count, selected_motifs, shuffle_scores, q_threshold);
  } else {
    xml_path = NULL;
    xml_output = NULL;
  }

  //Loop on motifs
  for (i = 1, link = linklst_first(motif_ll); link != NULL; ++i, link = link_next(link)) {
    char *motif;
    BINS_T *bins;
    AMA_LIST_T **motif_scores;

    // create one all inclusive bin 
    bins = bins_create(0, NULL);

    motif = link_item(link);
    motif_scores = (AMA_LIST_T**)hash_get_entry_value(hash_lookup_str(motif, motif_ht)); 
    
    if (status) fprintf(stderr, "Processing motif %s # %d of %d \n", motif, i, linklst_size(motif_ll));
    
    shuffler_identity(shuffler);
    calculate_scores_and_bin(goinfo_keys, gomap, motif_scores, species_count, FALSE, shuffler, bins);
    bins_sort(bins);

    //if (species_count >= 2) { //FIXME disabled while a decision is made on the validity of this optimisation
      if (status) fprintf(stderr, "Starting generation of null model scores\n");
      for (j = 1; j <= shuffle_scores; ++j) {
        shuffler_shuffle(shuffler);
        calculate_scores_and_bin(goinfo_keys, gomap, motif_scores, species_count, TRUE, shuffler, bins);
        if (status && (j % status_update_iter) == 0) {
          fprintf(stderr, "Generated %d%%\n", (int)(((double)j / (double)shuffle_scores) * 100.0));
        }
      }
    //} //FIXME disabled while a decision is made on the validity of this optimisation
    // calculate and output with qvalues
    calculate_qvalues_and_output(motif, bins, goinfo_ht, xml_output, text_only, q_threshold, species_count);
    // clean up
    bins_destroy(bins);
  }

  if (!text_only) {
    // finish output
    fprintf(xml_output,"</gomo>\n");
    fclose(xml_output);
  }

  if (status) fprintf(stderr, "Finished \n");

  if (!text_only) {
    // html files become to big for non-constrainted q-values
    if (q_threshold <= 0.5) {
      char *stylesheet_path, *html_path;
      stylesheet_path = make_path_to_file(ETC_DIR, HTML_STYLESHEET);
      html_path = make_path_to_file(output_dirname, HTML_FILENAME);
      if (file_exists(stylesheet_path)) {
        print_xml_filename_to_filename_using_stylesheet(
            xml_path,         /* path to XML input file IN */
            stylesheet_path,  /* path to MEME XSL stylesheet IN */
            html_path       /* path to HTML output file IN */
            );
      } else {
        if (verbosity >= NORMAL_VERBOSE) 
          fprintf(stderr, "Warning: could not find the stylesheet \"%s\" required for transformation of xml to html. Have you installed %s correctly?\n", 
            stylesheet_path, program_name);
      }
      free(stylesheet_path);
      free(html_path);
    } else {
      if (verbosity >= NORMAL_VERBOSE) 
        fprintf(stderr, "Warning: Reporting threshold for q-value (%g) is too permissive to allow HTML output due to size and memory constraints.\n",
            q_threshold);
    }
  }

  // measure time
  c1 = clock();
  if (verbosity >= NORMAL_VERBOSE) { // starting time
 	  fprintf(stderr,"cycles (CPU);            %ld cycles\n", (long) c1);
 	  fprintf(stderr,"elapsed CPU time:        %f seconds\n", (float) (c1 - c0)/CLOCKS_PER_SEC);
  }

  // clean up
  if (!text_only) free(xml_path);
  // destroy go map
  for (i = 0; i < arraylst_size(goinfo_keys); ++i) {
      free(gomap[i].seqnums);
  }
  free(gomap);
  // destroy motifs
  for (link = linklst_first(motif_ll); link != NULL; link = link_next(link)) {
    char *motif = link_item(link);
    if ((he = hash_lookup_str(motif, motif_ht)) != NULL) {
      AMA_LIST_T **motif_scores = (AMA_LIST_T**)hash_get_entry_value(he); 
      for (i = 0; i < species_count; ++i) {
        destroy_ama_score_list(motif_scores[i]);
      }
      free(motif_scores);
    } else {
      fprintf(stderr, "Warning, cleanup expected hash entry for %s in motif_ht\n", motif);
    }
    free(motif);
  }
  //destroy shuffler
  shuffler_destroy(shuffler);
  //destroy go terms and sequences
  for (i = 0; i <  arraylst_size(goinfo_keys); ++i) {
    if ((he = hash_lookup_str(arraylst_get(i, goinfo_keys), goinfo_ht)) != NULL) {
        GOINFO_T *info = (GOINFO_T*)hash_get_entry_value(he);
        free(info->desc);
        free(info);
    } else {
      fprintf(stderr, "Warning, cleanup expected hash entry for %s in goinfo_ht\n", (char*)arraylst_get(i, goinfo_keys));
    }
  }
  for (i = 0; i <  arraylst_size(seqid_keys); ++i) {
    if ((he = hash_lookup_str(arraylst_get(i, seqid_keys), seqid_ht)) != NULL) {
        int *seqno = (int*)hash_get_entry_value(he);
        free(seqno);
    } else {
      fprintf(stderr, "Warning, cleanup expected hash entry for %s in seqid_ht\n", (char*)arraylst_get(i, seqid_keys));
    }
  }
  arraylst_destroy(seqid_keys);
  hash_destroy(seqid_ht);
  arraylst_destroy(goinfo_keys);
  hash_destroy(goinfo_ht);
  //destroy file lists
  arraylst_destroy(ama_files);
  
  if (selected_motifs != NULL) arraylst_destroy(selected_motifs);
  return(0);
}


